#include "search.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <regex.h>

#include "buffer.h"
#include "utils.h"
#include "mess.h"

static void load_line(FILE *f, buffer *b)
{
  b->size = 0;
  while (1) {
    int c = fgetc(f);
    if (c == EOF || c == '\n') break;
    bchar(b, c);
  }
  bchar(b, 0);
}

void load_words(char *words, char *table, word_list *w)
{
  FILE *f;
  buffer b;
  int nwords;
  int i;
  unsigned long table_length;
  struct stat buf;

  init_buffer(&b, 64, 16);

  f = fopen(words, "rb"); if (f == NULL) { perror(words); exit(1); }
  load_line(f, &b);
  if (b.data[0] == 0) { printf("bad file: %s\n", words); exit(1); }
  sscanf(b.data, "%d", &nwords);
  if (nwords <= 0) { printf("bad file: %s\n", words); exit(1); }

  w->size = nwords;
  w->names = malloc(sizeof(char *) * nwords);
  if (w->names == NULL) { perror("malloc"); exit(1); }

  for (i = 0; i < nwords; i++) {
    load_line(f, &b);
    if (b.data[0] == 0) { printf("bad file: %s\n", words); exit(1); }
    w->names[i] = strdup(b.data);
    if (w->names[i] == NULL) { perror("strdup"); exit(1); }
  }
  fclose(f);

  close_buffer(&b);

  /* load table */
  w->docs = malloc(sizeof(docs_for_word) * nwords);
  if (w->docs == NULL) { perror("malloc"); exit(1); }
  f = fopen(table, "rb"); if (f == NULL) { perror(table); exit(1); }
  for (i = 0; i < nwords; i++) {
    unsigned char c[4];
    if (fread(c, 4, 1, f) != 1) { perror(table); exit(1); }
    w->docs[i].offset = c[0] | (c[1] << 8) | (c[2] << 16) | (c[3] << 24);
  }
  if (fstat(fileno(f), &buf) == -1) { perror(table); exit(1); }
  table_length = buf.st_size;
  /* set length */
  for (i = 0; i < nwords-1; i++)
    w->docs[i].size = (w->docs[i+1].offset - w->docs[i].offset) / 4;
  w->docs[i].size = (table_length - w->docs[i].offset) / 4;
  fclose(f);
}

int search_word(word_list *w, char *word)
{
  /* TODO: optimize (dichotomy) */
  int i;
  for (i = 0; i < w->size; i++)
    if (!strcmp(w->names[i], word)) return i;
  return -1;
}

static void search_regex(word_list *w, char *word, char *retmap, char *table)
{
  /* TODO: optimize, but how? */
  int i;
  regex_t reg;
  int err;

  if ((err = regcomp(&reg, word, REG_EXTENDED | REG_NOSUB)) !=0) {
    char r[2048];
    regerror(err, &reg, r, 2038);
    printf("regex error (regcomp) with '%s': %s\n", word, r);
    exit(1);
  }
  for (i = 0; i < w->size; i++)
    if (regexec(&reg, w->names[i], 0, 0, 0) == 0) {
      V("#MATCH word '%s' against pattern '%s'\n", w->names[i], word);
      feed_map(retmap, w, i, table);
    }
  regfree(&reg);
}

void feed_map(char *map, word_list *w, int n, char *table)
{
  FILE *f;
  unsigned char c[4];
  uint32_t pos;
  int i;

  f = fopen(table, "rb"); if (f == NULL) { perror(table); exit(1); }
  if (fseek(f, w->docs[n].offset, SEEK_SET)) { perror(table); exit(1); }
  for (i = 0; i < w->docs[n].size; i++) {
    if (fread(c, 4, 1, f) != 1) { perror(table); exit(1); }
    pos = c[0] | (c[1] << 8) | (c[2] << 16) | (c[3] << 24);
    if (pos >= 251287)
      { printf("%s: bad file (pos %u)\n", table, pos); exit(1); }
    map[pos] = 1;
  }
  fclose(f);
}

static char *getmap(void)
{
  char *ret = malloc(251287); if (!ret) { perror("malloc"); exit(1); }
  return ret;
}

unsigned char *tok(unsigned char *e, buffer *b, unsigned char *space)
{
  b->size = 0;
  /* skip spaces */
  while (*e && space[*e]) e++;
  /* sep */
  switch (*e) {
  case '[':
  case ']':
  case ',': bchar(b, *e); bchar(b, 0); return e+1;
  }
  /* word */
  while (*e && !space[*e] && *e != '[' && *e != ']' && *e != ',')
    { bchar(b, *e); e++; }
  bchar(b, 0);
  return e;
}

static word_list w;
static unsigned char space[256];
static unsigned char system_space[256];
static unsigned char upper[256];
static buffer b;
static char *table;
static char *_index;
static char *_cables;

static unsigned char *search_seq(unsigned char *r, char *retmap)
{
  int i;
  char *seqmap;
  buffer ns;
  buffer m;
  int count;
  regex_t reg;
  int err;

  init_buffer(&ns, 512, 512);
  seqmap = getmap();
  memset(retmap, 1, 251287);
  r = tok(r, &b, space); if (b.data[0] != '[') return NULL;
  while (1) {
    int n;
    memset(seqmap, 0, 251287);
    r = tok(r, &b, space); if (b.data[0] == 0) return NULL;
    if (!strcmp(b.data, "/REG")) {
      char sep = *r; if (sep == 0) return NULL;
      r++;
      b.size = 0;
      while (*r && *r != sep) { bchar(&b, *r); r++; }
      bchar(&b, 0);
      if (*r != sep) return NULL;
      r++;
      search_regex(&w, b.data, seqmap, table);
    } else {
      n = search_word(&w, b.data);
      if (n != -1) feed_map(seqmap, &w, n, table);
    }
    bstring(&ns, b.data);
    for (i = 0; i < 251287; i++) retmap[i] &= seqmap[i];
    r = tok(r, &b, space);
    if (b.data[0] == ']') break;
    if (b.data[0] != ',') return NULL;
    bchar(&ns, ' ');
  }
  bchar(&ns, 0);

  count = 0;
  for (i = 0; i < 251287; i++) if (retmap[i]) count++;
  V("#SEARCH regex '%s' in %d documents\n", ns.data, count);

  if ((err = regcomp(&reg, ns.data, REG_EXTENDED | REG_NOSUB)) !=0) {
    char r[2048];
    regerror(err, &reg, r, 2038);
    printf("regex error (regcomp) with '%s': %s\n", ns.data, r);
    exit(1);
  }

  init_buffer(&m, 32768, 16384);
  for (i = 0; i < 251287; i++) {
    if (!retmap[i]) continue;
    get_raw(_cables, _index, i, &m, system_space, upper);
    if (regexec(&reg, m.data, 0, 0, 0) != 0) retmap[i] = 0;
  }
  close_buffer(&m);

  close_buffer(&ns);
  free(seqmap);
  return r;
}

static unsigned char *_s(unsigned char *e, char *retmap)
{
  int i;
  unsigned char *r = tok(e, &b, space);
  memset(retmap, 0, 251287);
  if (b.data[0] == '[' || b.data[0] == ']' || b.data[0] == ',') return NULL;
  /* simple word (but regex) */
  if (b.data[0] != '/') {
    int n = search_word(&w, b.data);
    if (n != -1) feed_map(retmap, &w, n, table);
  } else
  /* and, or, not, seq */
  if (!strcmp(b.data, "/AND")) {
    char *andmap = getmap();
    memset(retmap, 1, 251287);
    r = tok(r, &b, space); if (b.data[0] != '[') return NULL;
    while (1) {
      r = _s(r, andmap); if (r == NULL) return NULL;
      for (i = 0; i < 251287; i++) retmap[i] &= andmap[i];
      r = tok(r, &b, space);
      if (b.data[0] == ']') break;
      if (b.data[0] != ',') return NULL;
    }
    free(andmap);
  } else
  if (!strcmp(b.data, "/OR")) {
    char *ormap = getmap();
    r = tok(r, &b, space); if (b.data[0] != '[') return NULL;
    while (1) {
      r = _s(r, ormap); if (r == NULL) return NULL;
      for (i = 0; i < 251287; i++) retmap[i] |= ormap[i];
      r = tok(r, &b, space);
      if (b.data[0] == ']') break;
      if (b.data[0] != ',') return NULL;
    }
    free(ormap);
  } else
  if (!strcmp(b.data, "/NOT")) {
    char *notmap = getmap();
    r = tok(r, &b, space); if (b.data[0] != '[') return NULL;
    r = _s(r, notmap); if (r == NULL) return NULL;
    for (i = 0; i < 251287; i++) retmap[i] = 1 - notmap[i];
    r = tok(r, &b, space);
    if (b.data[0] != ']') return NULL;
    free(notmap);
  } else
  if (!strcmp(b.data, "/REG")) {
    char sep = *r; if (sep == 0) return NULL;
    r++;
    b.size = 0;
    while (*r && *r != sep) { bchar(&b, *r); r++; }
    bchar(&b, 0);
    if (*r != sep) return NULL;
    r++;
    search_regex(&w, b.data, retmap, table);
  } else
  if (!strcmp(b.data, "/SEQ")) {
    r = search_seq(r, retmap); if (r == NULL) return NULL;
  } else
    return NULL;
  return r;
}

char *complex_search(char *cables, char *index_file, char *dict_file,
                     char *table_file, char *expression)
{
  unsigned char *x;
  char *retmap;

  table = table_file;
  _index = index_file;
  _cables = cables;

  spacify(system_space);
  spacify(space);
  space['/'] = 0;
  space['['] = 0;
  space[']'] = 0;
  space[','] = 0;
  upperify(upper);

  init_buffer(&b, 512, 512);
  load_words(dict_file, table_file, &w);

  retmap = getmap();

  x = _s((unsigned char *)expression, retmap);
  if (x) tok(x, &b, space);
  if (x == NULL || b.size != 1)
    { printf("bad search '%s'\n", expression); exit(1); }

  close_buffer(&b);

  return retmap;
}
